module Structural where
open import Utils
open import Term
open import Value
open import Normal
open import BigStep

mutual
  ev⁼ : forall {Γ}{Δ}{σ}(t : Tm Δ σ)(vs : Env Γ Δ){v : Val Γ σ} -> 
        t at vs ⇓ v -> Σ (Val Γ σ) \v'  -> v' == v
  ev⁼ ø          (_ ◂ v)  rø            = sig v  refl⁼  
  ev⁼ (λ t)      vs       rλ            = sig (λv t vs) refl⁼ 
  ev⁼ (t $ u)    vs       (r$ p p' p'') with ev⁼ t vs p  | ev⁼ u vs p'
  ev⁼ (t $ u)    vs       (r$ p p' p'')    | sig f refl⁼ | sig a refl⁼ = 
    f $$⁼ a & p'' 
  ev⁼ (t [ ts ]) vs       (r[] p p') with evˢ⁼ ts vs p
  ev⁼ (t [ ts ]) vs       (r[] p p')    | sig vs' refl⁼ = ev⁼ t vs' p' 
  ev⁼ Z          vs       rZ            = sig Zv refl⁼
  ev⁼ (S n)      vs       (rS p)     with ev⁼ n vs p
  ev⁼ (S n)      vs       (rS p)        | sig v refl⁼ = sig (Sv v) refl⁼
  ev⁼ (pr n f z) vs (rp p q r s) with ev⁼ n vs p | ev⁼ f vs q | ev⁼ z vs r
  ev⁼ (pr n f z) vs (rp p q r s) | sig v refl⁼ | sig v' refl⁼ | sig v'' refl⁼ =
     prv v v' v'' s

  prv : forall {Γ}{σ}(n : Val Γ N)(f : Val Γ (N → σ → σ))(z : Val Γ σ) ->
        {v : Val Γ σ} -> pr⇓ n f z v -> Σ (Val Γ σ) \v' -> v' == v
  prv Zv     _ z rpZ = sig z refl⁼
  prv (Sv n) f z (rpS p q r) with f $$⁼ n & p  | prv n f z q 
  prv (Sv n) f z (rpS p q r)    | sig fn refl⁼ | sig pn refl⁼ = 
    fn $$⁼ pn & r
  prv (nev n) f z rpn = sig (nev (npr n f z)) refl⁼   

  _$$⁼_&_ : {Γ : Cn} -> {σ τ : Ty} -> (f : Val Γ (σ → τ)) -> (a : Val Γ σ) ->
            {v : Val Γ τ} -> f $ a ⇓ v -> Σ (Val Γ τ) \v' -> v' == v
  (λv t vs) $$⁼ v & (r$λ p) = ev⁼ t (vs ◂ v) p  
  (nev n)   $$⁼ v & r$n     = sig (nev (app n v)) refl⁼  

  evˢ⁼ : {B Γ Δ : Cn} -> (ts : Tmˢ Γ Δ) -> (vs : Env B Γ) -> 
        {ws : Env B Δ} -> ts atˢ vs ⇓ ws -> Σ (Env B Δ) \ws' -> ws' == ws
  evˢ⁼ id        vs       rid       = sig vs refl⁼ 
  evˢ⁼ (ts ◃ t) vs (r◃ p p') with evˢ⁼ ts vs p | ev⁼ t vs p'
  evˢ⁼ (ts ◃ t) vs (r◃ p p')    | sig vs' refl⁼  | sig v refl⁼ = 
     sig (vs' ◂ v) refl⁼  
  evˢ⁼ (ts ○ us) vs       (r○ p p') with evˢ⁼ us vs p 
  evˢ⁼ (ts ○ us) vs       (r○ p p')    | sig ws refl⁼ = evˢ⁼ ts ws p' 

mutual
  quote⁼ : forall {Γ}{σ}(v : Val Γ σ){n : Nf Γ σ} -> q v ⇓ n -> 
           Σ (Nf Γ σ) \n' -> n' == n
  quote⁼ {Γ}{σ → τ} f (q→ p p') with wkv σ f $$⁼ nev (var vZ) & p
  ... | sig v refl⁼ with quote⁼ v p'
  ... | sig n refl⁼ = sig (λn n) refl⁼
  quote⁼ {Γ}{N} (nev n) (qn p) with quoteⁿ⁼ n p
  ... | sig n' refl⁼ = sig (nen n') refl⁼
  quote⁼ {Γ}{N} Zv qZ = sig Zn refl⁼
  quote⁼ {Γ}{N} (Sv n) (qS p) with quote⁼ n p 
  ... | sig n' refl⁼ = sig (Sn n') refl⁼

  quoteⁿ⁼ : forall {Γ}{σ}(n : Ne Val Γ σ){n' : Ne Nf Γ σ} -> qⁿ n ⇓ n' -> 
           Σ (Ne Nf Γ σ) \n'' -> n'' == n'
  quoteⁿ⁼ (var v) qv = sig (var v) refl⁼
  quoteⁿ⁼ (app n v) (qa p p') with quoteⁿ⁼ n p | quote⁼ v p'
  ... | sig n' refl⁼ | sig n'' refl⁼ = sig (app n' n'') refl⁼
  quoteⁿ⁼ (npr n f z) (qpr a b c) with quoteⁿ⁼ n a | quote⁼ f b | quote⁼ z c
  ... | sig n' refl⁼ | sig f' refl⁼ | sig z' refl⁼ = sig (npr n' f' z') refl⁼

nf⁼ : forall {Γ}{σ}(t : Tm Γ σ){n : Nf Γ σ} -> nf t ⇓ n -> 
      Σ (Nf Γ σ) \n' -> n' == n
nf⁼ t (nf⇓ (sig v (pa p p'))) with ev⁼ t ide p
... | sig .v refl⁼ = quote⁼ v p'