module Term where
open import Utils

infixr 50 _→_
infixl 50 _$_
infixl 50 _○_
infixr 50 _◃_

data Ty : Set where
  N   : Ty
  _→_ : Ty -> Ty -> Ty

data Cn : Set where
  ε   : Cn
  _,_ : Cn -> Ty -> Cn

mutual
  data Tm : Cn -> Ty -> Set where
    ø    : forall {Γ}{σ} -> Tm (Γ , σ) σ
    λ    : forall {Γ}{σ τ} -> Tm (Γ , σ) τ -> Tm Γ (σ → τ)
    _$_  : forall {Γ}{σ}{τ} -> Tm Γ (σ → τ) -> Tm Γ σ -> Tm Γ τ
    _[_] : forall {Γ}{Δ}{σ} -> Tm Δ σ -> Tmˢ Γ Δ -> Tm Γ σ
    Z    : forall {Γ} -> Tm Γ N
    S    : forall {Γ} -> Tm Γ N -> Tm Γ N
    pr   : forall {Γ}{σ} -> Tm Γ N -> Tm Γ (N → σ → σ) -> Tm Γ σ -> Tm Γ σ

  data Tmˢ : Cn -> Cn -> Set where
    id  : forall {Γ}         -> Tmˢ Γ Γ
    _◃_ : forall {Γ}{Δ}{σ}   -> Tmˢ Γ Δ -> Tm Γ σ -> Tmˢ Γ (Δ , σ)
    ↑   : forall {Γ}(σ : Ty) -> Tmˢ (Γ , σ) Γ
    _○_ : forall {B}{Γ}{Δ}   -> Tmˢ Γ Δ -> Tmˢ B Γ -> Tmˢ B Δ

infixl 50 _++_
 
_++_ : Cn -> Cn -> Cn
Γ ++ ε       = Γ
Γ ++ (Δ , σ) = (Γ ++ Δ) , σ  

assoc : (B Γ Δ : Cn) -> (B ++ Γ ++ Δ) == (B ++ (Γ ++ Δ))
assoc B Γ ε       = refl⁼ 
assoc B Γ (Δ , σ) = resp (\B -> B , σ) (assoc B Γ Δ)

↑++ : forall {Γ} Δ -> Tmˢ (Γ ++ Δ) Γ
↑++ ε = id
↑++ (Δ , σ) = ↑++ Δ ○ ↑ σ

mutual
  data _≡_ : forall {Γ}{σ}(t t' : Tm Γ σ) -> Set where
    -- equivalence
    refl  : forall {Γ}{σ}{t : Tm Γ σ}        -> t ≡ t
    sym   : forall {Γ}{σ}{t t' : Tm Γ σ}     -> t ≡ t' -> t' ≡ t
    trans : forall {Γ}{σ}{t t' t'' : Tm Γ σ} -> t ≡ t' -> t' ≡ t'' -> t ≡ t''
    
    -- congruence
    λ≡    : forall {Γ}{σ}{τ}{t t' : Tm (Γ , σ) τ} -> t ≡ t' -> λ t ≡ λ t'
    $≡    : forall {Γ}{σ}{τ}{t t' : Tm Γ (σ → τ)} -> t ≡ t' -> 
            {u u' : Tm Γ σ} -> u ≡ u' -> t $ u ≡ t' $ u'
    []≡   : forall {Γ}{Δ}{σ}{t t' : Tm Δ σ} -> t ≡ t' -> {ts ts' : Tmˢ Γ Δ} ->
            ts ≡ˢ ts' ->  t [ ts ] ≡  t' [ ts' ]
    S≡    : forall {Γ}{n n' : Tm Γ N} -> n ≡ n' -> S n ≡ S n'
    pr≡ : forall {Γ}{σ}{n n' : Tm Γ N} -> n ≡ n' -> 
            {f f' : Tm Γ (N → σ → σ)} -> f ≡ f' -> {z z' : Tm Γ σ} -> z ≡ z' ->
            pr n f z ≡ pr n' f' z'

    -- computation
    ø≡ : forall {Γ}{Δ}{σ}{ts : Tmˢ Γ Δ}{t : Tm Γ σ} ->  ø [ ts ◃ t ] ≡ t
    [][]≡ : forall {B}{Γ}{Δ}{σ}{t : Tm Δ σ}{ts : Tmˢ Γ Δ}{us : Tmˢ B Γ} ->
             t [ ts ] [ us ] ≡  t [ ts ○ us ]
    $[]≡ : forall {Γ}{Δ}{σ}{τ}{t : Tm Δ (σ → τ)}{u : Tm Δ σ}{ts : Tmˢ Γ Δ} ->
            t $ u [ ts ] ≡ (t [ ts ]) $ (u [ ts ])
    λ[]≡ : forall {Γ}{Δ}{σ}{τ}{t : Tm (Δ , σ) τ}{ts : Tmˢ Γ Δ} ->
            λ t [ ts ] ≡ λ ( t  [ (ts ○ ↑ σ) ◃ ø ])
    Z[]≡ : forall {Γ Δ}{ts : Tmˢ Γ Δ} -> Z [ ts ] ≡ Z {Γ}
    S[]≡ : forall {Γ Δ}{ts : Tmˢ Γ Δ}{n : Tm Δ N} -> S n [ ts ] ≡ S (n [ ts ])
    pr[]≡ : forall {Γ Δ}{σ}{t : Tm Δ N}{f : Tm Δ (N → σ → σ)}{z : Tm Δ σ} ->
            {ts : Tmˢ Γ Δ} -> 
            pr t f z [ ts ] ≡ pr (t [ ts ]) (f [ ts ]) (z [ ts ]) 
    prZ≡ : forall {Γ}{σ}{f : Tm Γ (N → σ → σ)}{z : Tm Γ σ} ->
           pr Z f z ≡ z
    prS≡ : forall {Γ}{σ}{f : Tm Γ (N → σ → σ)}{z : Tm Γ σ}{n : Tm Γ N} ->
           pr (S n) f z ≡ f $ n $ pr n f z
    βσ : forall {Γ Δ}{σ τ}{t : Tm (Δ , σ) τ}{ts : Tmˢ Γ Δ}{u : Tm Γ σ} ->
         (λ t [ ts ]) $ u ≡ t [ ts ◃ u ]
    η : forall {Γ}{σ τ}{f : Tm Γ (σ → τ)} -> f ≡ λ ((f [ ↑ σ ]) $ ø)
    id≡ : forall {Γ}{σ}{t : Tm Γ σ} -> t ≡ t [ id ]

  data _≡ˢ_ : forall {Γ}{Δ} -> Tmˢ Γ Δ -> Tmˢ Γ Δ -> Set where
    -- equivalence
    reflˢ  : forall {Γ}{Δ}{ts : Tmˢ Γ Δ}          -> ts ≡ˢ ts
    symˢ   : forall {Γ}{Δ}{ts ts' : Tmˢ Γ Δ}      -> ts ≡ˢ ts' -> ts' ≡ˢ ts
    transˢ : forall {Γ}{Δ}{ts ts' ts'' : Tmˢ Γ Δ} -> ts ≡ˢ ts' -> 
             ts' ≡ˢ ts'' -> ts ≡ˢ ts''

    -- congruence
    ◃≡ : forall {Γ}{Δ}{σ}{ts ts' : Tmˢ Γ Δ} -> ts ≡ˢ ts' -> {t t' : Tm Γ σ} ->
         t ≡ t' -> ts ◃ t ≡ˢ ts' ◃ t'
    ○≡ : forall {B}{Γ}{Δ}{ts ts' : Tmˢ Γ Δ} -> ts ≡ˢ ts' -> 
         {us us' : Tmˢ B Γ} -> us ≡ˢ us' -> ts ○ us ≡ˢ ts' ○ us' 

    -- computation
    ◃id≡ˢ : forall {Γ}{σ} -> id {Γ , σ} ≡ˢ ((id ○ ↑ σ) ◃ ø)
    ↑◃≡ˢ : forall {Γ}{Δ}{σ}{ts : Tmˢ Γ Δ}{t : Tm Γ σ} -> ↑ σ ○ (ts ◃ t) ≡ˢ ts
    id≡ˢ : forall {Γ}{Δ}{ts : Tmˢ Γ Δ} -> id ○ ts ≡ˢ ts
    ○○≡ˢ : forall {A}{B}{Γ}{Δ}{ts : Tmˢ Γ Δ}{us : Tmˢ B Γ}{vs : Tmˢ A B} ->
           ts ○ us ○ vs ≡ˢ ts ○ (us ○ vs)
    ○◃≡ˢ : forall {B}{Γ}{Δ}{σ}{ts : Tmˢ Γ Δ}{t : Tm Γ σ}{us : Tmˢ B Γ} ->
           (ts ◃ t) ○ us ≡ˢ (ts ○ us) ◃  ( t [ us ])

mkid : forall Γ -> Tmˢ Γ ε
mkid ε       = id
mkid (Γ , σ) = mkid Γ ○ ↑ σ  
